#!/usr/bin/env python3
####################################################################
#   Copyright (C) 2020-2021 Luis Falcon <falcon@gnuhealth.org>
#   Copyright (C) 2020-2021 GNU Solidario <health@gnusolidario.org>
#   License: GPL v3+
#   Please read the COPYRIGHT and LICENSE files of the package
####################################################################

from mygnuhealth import about
from PySide2.QtCore import QObject, Property


class GHAbout(QObject):

    def aboutmygh(self):
        version = about.__version__
        author = about.__author__
        homepage = about.__homepage__
        copyright = about.__copyright__
        license = about.__license__
        info_email = about.__email__
        thanks = about.__thanks__

        credits = {'version': version, 'author': author,
                   'homepage': homepage, 'copyright': copyright,
                   'license': license, 'info_email': info_email,
                   'thanks': thanks}

        return credits

    # Expose the aboubt / credits dictionary to QML
    credits = Property("QVariant", aboutmygh, constant=True)



#!/usr/bin/env python3
##############################################################################
#
#    MyGNUHealth : Mobile and Desktop PHR node for GNU Health
#
#           MyGNUHealth is part of the GNU Health project
#
##############################################################################
#
#    GNU Health: The Libre Digital Health Ecosystem
#    Copyright (C) 2008-2021 Luis Falcon <falcon@gnuhealth.org>
#    Copyright (C) 2011-2021 GNU Solidario <health@gnusolidario.org>
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
##############################################################################

import datetime
import dateutil.parser
import os
from tinydb import TinyDB
from mygnuhealth.myghconf import dbfile, bolfile

''' Import the Natural variants database from the data
directory, relative to the this mygnuhealth module
The datafile is loaded into the TinyDB "vardb" variable
'''
moddir = os.path.dirname(os.path.abspath(__file__))
varfile = os.path.join(moddir, 'data/variants.db')
vardb = TinyDB(varfile, access_mode='r')


def datefromisotz(isotz):
    """This method is compatible with Python 3.6 .
    datetime fromisoformat is not present until Python 3.7
    """

    if isotz:
        return dateutil.parser.parse(isotz)


def get_user_profile(db):
    """Retrieves the user profile (DoB, sex, height ...)"""

    profile_table = db.table('profile')
    profile = None
    # Credentials table holds a singleton, so only one record
    if (len(profile_table) > 0):
        profile = profile_table.all()[0]
    return profile


def get_personal_key(db):
    """Retrieves the user personal key"""

    credentials_table = db.table('credentials')
    # Credentials table holds a singleton, so only one record
    personal_key = credentials_table.all()[0]['personal_key']
    return personal_key.encode()


def get_federation_account():
    """Retrieves the user GH Federation account, if any."""

    db = TinyDB(dbfile)
    federation_table = db.table('federation')
    fedacct = None
    if (len(federation_table) > 0):
        # Federation Account table holds a singleton, so only one record
        res = federation_table.all()[0]
        if 'federation_account' in res.keys():
            fedacct = res['federation_account']
            print("Fed account ", fedacct)
    return fedacct


def check_date(date):
    """ Verifies that the entered date is valid"""
    year, month, day = date
    try:
        datetime.date(year, month, day)
        return True
    except ValueError:
        print("Invalid date")
        return False


class PageOfLife():
    """
    Page of Life
    The basic shema of PoL from GH Federation HIS, used  by Thalamus

    Attributes
    ----------
        boldb: TinyDB instance
            The book of life DB. It contains all the Pages of Life created
            by the user.

        pol_model : dict
            Dictionary holding the schema of the GNU Healtth Federation
            Health Information System  database

        medical_context: In a page of life, when the medical domain is chosen,
            the user can choose

        social_context: The different contexts within the Social domain.

        Methods:
        --------
            create_pol: Creates a Page of Life associated the event / reading

    """
    boldb = TinyDB(bolfile)

    pol_model = dict.fromkeys([
        'book', 'page_date', 'age', 'domain', 'relevance', 'privacy',
        'context', 'measurements', 'genetic_info', 'summary', 'info',
        'node', 'author', 'author_acct', 'fsynced'
        ])

    pol_domain = [{'value': 'medical', 'text': 'Medical'},
                  {'value': 'social', 'text': 'Social'},
                  {'value': 'biographical', 'text': 'Biographical'},
                  {'value': 'lifestyle', 'text': 'Lifestyle'},
                  {'value': 'other', 'text': 'Other'}
                  ]

    medical_context = [
        {'value': 'health_condition', 'text': 'Health Condition'},
        {'value': 'encounter', 'text': 'Encounter'},
        {'value': 'procedure', 'text': 'Procedure'},
        {'value': 'self_monitoring', 'text': 'Self monitoring'},
        {'value': 'immunization', 'text': 'Immunization'},
        {'value': 'prescription', 'text': 'Prescription'},
        {'value': 'surgery', 'text': 'Surgery'},
        {'value': 'hospitalization', 'text': 'Hospitalization'},
        {'value': 'lab', 'text': 'Lab test'},
        {'value': 'dx_imaging', 'text': 'Dx Imaging'},
        {'value': 'genetics', 'text': 'Genetics'},
        {'value': 'family', 'text': 'Family history'},
        ]

    social_context = [
        {'value': 'social_gradient', 'text': 'Social Gradient / Equity'},
        {'value': 'stress', 'text': 'Stress'},
        {'value': 'early_life_development', 'text': 'Early life development'},
        {'value': 'social_exclusion', 'text': 'Social exclusion'},
        {'value': 'working_conditions', 'text': 'Working conditions'},
        {'value': 'education', 'text': 'Education'},
        {'value': 'physical_environment', 'text': 'Physical environment'},
        {'value': 'unemployment', 'text': 'Unemployment'},
        {'value': 'social_support', 'text': 'Social Support'},
        {'value': 'addiction', 'text': 'Addiction'},
        {'value': 'food', 'text': 'Food'},
        {'value': 'transportation', 'text': 'Transportation'},
        {'value': 'health_services', 'text': 'Health services'},
        {'value': 'family_functionality', 'text': 'Family functionality'},
        {'value': 'family_violence', 'text': 'Family violence'},
        {'value': 'bullying', 'text': 'Bullying'},
        {'value': 'war', 'text': 'War'},
        {'value': 'misc', 'text': 'Misc'},
        ]

    lifestyle_context = [
        {'value': 'physical_activity', 'text': 'Physical Activity'},
        {'value': 'nutrition', 'text': 'Nutrition'},
        {'value': 'sleep', 'text': 'Sleep'},
        ]

    biographical_context = [
        {'value': 'birth', 'text': 'Birth'},
        {'value': 'death', 'text': 'Death'},
        {'value': 'misc', 'text': 'Misc'}
        ]

    other_context = [
        {'value': 'misc', 'text': 'Misc'},
        ]

    def create_pol(self, pol_vals):
        """Creates a Page of Life associated to the reading

        Parameters
        ----------
        pol_vals: Takes all the values from the page of life, which is a
        dictionary. Some of them:
            domain: the domain (medical, psycho, social)
            context: the context within a domain (possible contexts are listed
                in the core module.
            genetic_info: variant, rsref, protein, gene, aa_change
            measurements: blood pressure, Osat, temp, heart & resp frequency,..
            summary: Short description / title of the page
            info: Extended information related to this page of life.
        """

        node = "mygnuhealth"  # The node name is generic. "mygnuhealth"
        fed_acct = get_federation_account()
        poltable = self.boldb.table('pol')
        page_of_life = self.pol_model
        domain = pol_vals['domain']
        context = pol_vals['context']

        if (fed_acct):
            #  If the Federation account does not exist, it will be
            #  a private entry, not linked to a book or author
            #  and it won't be shared in the GNU Health Federation

            print("Retrieved Federation Account: ", fed_acct)
            page_of_life['book'] = fed_acct
            page_of_life['author'] = fed_acct
            page_of_life['author_acct'] = fed_acct

        page_of_life['node'] = node
        page_of_life['page'] = pol_vals['page']
        page_of_life['page_date'] = pol_vals['page_date']
        page_of_life['domain'] = domain

        page_of_life['context'] = context

        if ('relevance' in pol_vals.keys()):
            page_of_life['relevance'] = pol_vals['relevance']

        if ('privacy' in pol_vals.keys()):
            page_of_life['privacy'] = pol_vals['privacy']

        if ('genetic_info' in pol_vals.keys()):
            page_of_life['genetic_info'] = pol_vals['genetic_info'] or ''

        if ('measurements' in pol_vals.keys()):
            page_of_life['measurements'] = pol_vals['measurements']
        if ('summary' in pol_vals.keys()):
            page_of_life['summary'] = pol_vals['summary'] or ''
        if ('info' in pol_vals.keys()):
            page_of_life['info'] = pol_vals['info'] or ''

        # The fsync key reflects whether the page has been sent to the
        # GNU Health Federation HIS (Health Information System)
        page_of_life['fsynced'] = False
        # create the new PoL entry

        print("New Page of Life:", page_of_life)
        data = page_of_life
        poltable.insert(data)

        # Sample measurements keys accepted by Thalamus / GH Federation HIS
        #  {'bp': {'systolic': 116, 'diastolic': 79}, 't': 36.0, 'hr': 756, '
        #    rr': 16,
        #  'osat': 99, 'wt': 68.0, 'ht': 168.0, 'bmi': 24.09, 'bg': 116}



####################################################################
#   Copyright (C) 2020-2021 Luis Falcon <falcon@gnuhealth.org>
#   Copyright (C) 2020-2021 GNU Solidario <health@gnusolidario.org>
#   License: GPL v3+
#   Please read the COPYRIGHT and LICENSE files of the package
####################################################################

import datetime
from uuid import uuid4
from PySide2.QtCore import QObject, Signal, Slot
from tinydb import TinyDB
from mygnuhealth.myghconf import dbfile
from mygnuhealth.core import PageOfLife


class BloodPressure(QObject):

    db = TinyDB(dbfile)

    def insert_values(self, systolic, diastolic, heart_rate):
        blood_pressure = self.db.table('bloodpressure')
        hr = self.db.table('heart_rate')
        current_date = datetime.datetime.now().isoformat()
        bpmon = hrmon = False  # Init to false the bp and hr monitoring process
        domain = 'medical'
        context = 'self_monitoring'

        if (systolic > 0) and (diastolic > 0):
            bpmon = True
            bp_event_id = str(uuid4())
            synced = False
            blood_pressure.insert({'timestamp': current_date,
                                   'event_id': bp_event_id,
                                   'synced': synced,
                                   'systolic': systolic,
                                   'diastolic': diastolic})

            print("Saved blood pressure", bp_event_id, synced, systolic,
                  diastolic, current_date)

        if heart_rate > 0:
            hrmon = True
            hr_event_id = str(uuid4())
            synced = False
            hr.insert({'timestamp': current_date,
                       'event_id': hr_event_id,
                       'synced': synced,
                       'heart_rate': heart_rate})

            print("Saved Heart rate", hr_event_id, synced,
                  heart_rate, current_date)

        if (bpmon or hrmon):
            # This block is related to the Page of Life creation
            if (bpmon and hrmon):
                # Group both HR and BP monitors in one PoL if both readings
                # where taken at the same moment / device
                # The event_id will be unique
                event_id = str(uuid4())
                monitor_readings = [
                    {'bp': {'systolic': systolic, 'diastolic': diastolic}},
                    {'hr': heart_rate}
                    ]
            elif (bpmon and not hrmon):
                event_id = bp_event_id
                monitor_readings = [
                    {'bp': {'systolic': systolic, 'diastolic': diastolic}},
                    ]
            else:
                event_id = hr_event_id
                monitor_readings = [{'hr': heart_rate}]

            pol_vals = {
                'page': event_id,
                'page_date': current_date,
                'domain': domain,
                'context': context,
                'measurements': monitor_readings
                }

            # Create the Page of Life associated to this reading
            PageOfLife.create_pol(PageOfLife, pol_vals)

    @Slot(int, int, int)
    def getvals(self, *args):
        self.insert_values(*args)
        self.setOK.emit()

    # Signal to emit to QML if the blood pressure values were stored correctly
    setOK = Signal()



####################################################################
#   Copyright (C) 2020-2021 Luis Falcon <falcon@gnuhealth.org>
#   Copyright (C) 2020-2021 GNU Solidario <health@gnusolidario.org>
#   License: GPL v3+
#   Please read the COPYRIGHT and LICENSE files of the package
####################################################################

from PySide2.QtCore import QObject, Signal, Slot, Property
from tinydb import TinyDB, Query
from uuid import uuid4
from mygnuhealth.myghconf import dbfile
from mygnuhealth.core import check_date, PageOfLife, vardb

import datetime


class PoL(QObject):
    """This class creates a new page in the user's Book of Life

        Attributes:
        -----------
            wrongDate: Signal to emit when an invalid date is found
            todayDate: Property with current date

        Methods:
        --------
            get_domains: Returns main domains (medical, social, biographical..)

    """

    def __init__(self):
        QObject.__init__(self)
        self.db = TinyDB(dbfile)
        self.domain = 'medical'
        self.pol_context = ''
        self.rsinfo = {}

    def get_rsinfo(self):
        return self.rsinfo

    def get_domains(self):
        """ Return the domains to be used in the QML form
            We use value and text keys to match QML Combobox"""
        self.domain = PageOfLife.pol_domain
        return self.domain

    def get_contexts(self):
        """ Return the domains to be used in the QML form
            We use value and text keys to match QML Combobox"""
        if self.domain == 'social':
            self.pol_context = PageOfLife.social_context
        if self.domain == 'medical':
            self.pol_context = PageOfLife.medical_context
        if self.domain == 'lifestyle':
            self.pol_context = PageOfLife.lifestyle_context
        if self.domain == 'biographical':
            self.pol_context = PageOfLife.biographical_context
        if self.domain == 'other':
            self.pol_context = PageOfLife.other_context

        return self.pol_context

    @Signal
    def domainChanged(self):
        pass

    @Signal
    def rsChanged(self):
        pass

    @Slot(str)
    def update_context(self, domain):
        """ Set the value of the domain from the QML selection"""
        if domain == 'social':
            self.pol_context = PageOfLife.social_context
        if domain == 'medical':
            self.pol_context = PageOfLife.medical_context
        if domain == 'lifestyle':
            self.pol_context = PageOfLife.lifestyle_context
        if domain == 'biographical':
            self.pol_context = PageOfLife.biographical_context
        if domain == 'other':
            self.pol_context = PageOfLife.other_context
        # Emit the change of domain, so it updates the context
        self.domainChanged.emit()

        return self.pol_context

    def get_date(self):
        """
        Returns the date packed into an array (day,month,year)
        """
        rightnow = datetime.datetime.now()
        dateobj = []
        dateobj.append(rightnow.day)
        dateobj.append(rightnow.month)
        dateobj.append(rightnow.year)
        dateobj.append(rightnow.hour)
        dateobj.append(rightnow.minute)
        return dateobj

    def new_page(self, data):
        page_id = str(uuid4())

        pol_vals = {
            'page': page_id,
            'page_date': data['page_date'],
            'domain': data['domain'],
            'context': data['context'],
            'relevance': data['relevance'],
            'privacy': data['privacy'],
            'summary': data['summary'],
            'info': data['info']
            }
        if (data['context'] == 'genetics'):
            pol_vals.update({'genetic_info': data['genetic_info']})

        PageOfLife.create_pol(PageOfLife, pol_vals)

    @Slot(str)
    def checkSNP(self, rs):
        Rsnp = Query()
        self.rsinfo = {}
        res = vardb.search(Rsnp.dbsnp == rs)
        if len(res) > 0:
            res = res[0]
            self.rsinfo = {
                    'rsid': res['dbsnp'],
                    'gene': res['gene'],
                    'aa_change': res['aa_change'],
                    'variant': res['variant'],
                    'protein': res['protein'],
                    'significance': res['category'],
                    'disease': res['disease']
                    }

        else:
            self.rsinfo = {}
            self.rsNotfound.emit()


        return (self.rsinfo)

    @Slot(list, str, str, str, bool, list, str, str)
    def createPage(self, page_date, domain, context, relevance, private_page,
                   genetic_info, summary, info):
        # Retrieves the information from the initialization form
        # Creates the page from the information on the form
        if (page_date):
            if (check_date(page_date[:3])):
                # Sets the page of life date and time
                year, month, day, hour, minute = page_date
                daterp = str(datetime.datetime(year, month, day, hour, minute))
                page = {'page_date': daterp,
                        'domain': domain,
                        'context': context,
                        'relevance': relevance,
                        'privacy': private_page,
                        'summary': summary,
                        'info': info
                        }
                if (context == 'genetics'):
                    rsinfo = {
                        'rsid': genetic_info[0],
                        'gene': genetic_info[1],
                        'aa_change': genetic_info[2],
                        'variant': genetic_info[3],
                        'protein': genetic_info[4],
                        'significance': genetic_info[5],
                        'disease': genetic_info[6]
                        }
                    page.update({'genetic_info': rsinfo})
                self.new_page(page)
            else:
                self.wrongDate.emit()

        self.createSuccess.emit()

    # Properties
    todayDate = Property("QVariantList", get_date, constant=True)
    poldomain = Property("QVariantList", get_domains, constant=True)
    polcontext = Property("QVariantList", get_contexts, notify=domainChanged)
    polrs = Property("QVariant", get_rsinfo, notify=rsChanged)

    # Signals
    createSuccess = Signal()
    wrongDate = Signal()
    rsNotfound = Signal()



####################################################################
#   Copyright (C) 2020-2021 Luis Falcon <falcon@gnuhealth.org>
#   Copyright (C) 2020-2021 GNU Solidario <health@gnusolidario.org>
#   License: GPL v3+
#   Please read the COPYRIGHT and LICENSE files of the package
####################################################################

import io
from PySide2.QtCore import QObject, Signal, Property
from tinydb import TinyDB
import matplotlib.pyplot as plt
import base64
from mygnuhealth.core import datefromisotz
from mygnuhealth.myghconf import dbfile


class GHLifestyle(QObject):

    def __init__(self):
        QObject.__init__(self)

        self.current_pa = ""

    db = TinyDB(dbfile)

    # PHYSICAL ACTIVITY
    def read_physical_activity(self):
        # Retrieve the physical activity history
        physical_activity = self.db.table('physicalactivity')
        pahist = physical_activity.all()
        return pahist

    def getPA(self):
        # Extracts the latest readings from Physical Activity table
        pahist = self.read_physical_activity()
        paobj = ['', '', '', '']  # Init to empty string to avoid undefined val
        if pahist:
            pa = pahist[-1]  # Get the latest (newest) record

            # dateobj =  datetime.datetime.fromisoformat(bp['timestamp'])
            dateobj = datefromisotz(pa['timestamp'])
            date_repr = dateobj.strftime("%a, %b %d '%y - %H:%M")

            paobj = [str(date_repr), str(pa['aerobic']),
                     str(pa['anaerobic']), str(pa['steps'])]

        return paobj

    def paplot(self):
        # Retrieves all the history and packages into an array.
        pahist = self.read_physical_activity()
        pa_aerobic = []
        pa_anaerobic = []
        pa_steps = []
        pa_date = []
        for element in pahist:

            dateobj = datefromisotz(element['timestamp'])

            # Only print one value per day to avoid artifacts in plotting.
            pa_date.append(dateobj)
            pa_aerobic.append(element['aerobic'])
            pa_anaerobic.append(element['anaerobic'])
            pa_steps.append(element['steps'])

        fig, axs = plt.subplots(3)

        # Plot aerobic, anaerobic and steps
        axs[0].plot(pa_date, pa_aerobic)
        axs[1].plot(pa_date, pa_anaerobic, color='teal')
        axs[2].plot(pa_date, pa_steps)

        axs[0].set_ylabel('Aerobic', size=13)
        axs[1].set_ylabel('Anaerobic', size=13)
        axs[2].set_ylabel('Steps', size=13)

        fig.autofmt_xdate()
        fig.suptitle("Time (minutes)", size=20)
        holder = io.BytesIO()
        fig.savefig(holder, format="svg")
        decoded_holder = base64.b64encode(holder.getvalue()).decode()
        image = f"data:image/svg+xml;base64,{decoded_holder}"

        holder.close()
        return image

    def setPA(self, pa):
        self.current_pa = pa
        # Call the notifying signal
        self.paChanged.emit()

    paChanged = Signal()

    # NUTRITION
    def read_nutrition(self):
        # Retrieve the nutrition history
        nutrition = self.db.table('nutrition')
        nutrihist = nutrition.all()
        return nutrihist

    def getNutrition(self):
        # Extracts the latest readings from Physical Activity table
        nutrihist = self.read_nutrition()
        # Init to empty string to avoid undefined val
        nutriobj = ['', '', '', '', '']
        if nutrihist:
            nutri = nutrihist[-1]  # Get the latest (newest) record

            dateobj = datefromisotz(nutri['timestamp'])
            date_repr = dateobj.strftime("%a, %b %d '%y - %H:%M")

            nutriobj = [str(date_repr), str(nutri['calmorning']),
                        str(nutri['calafternoon']),
                        str(nutri['calevening']),
                        str(nutri['caltotal'])
                        ]

        return nutriobj

    def nutriplot(self):
        # Retrieves all the history and packages into an array.
        nutrihist = self.read_nutrition()
        nutri_calmorning = []
        nutri_calafternoon = []
        nutri_calevening = []
        nutri_caltotal = []
        nutri_date = []
        for element in nutrihist:

            dateobj = datefromisotz(element['timestamp'])

            # Only print one value per day to avoid artifacts in plotting.
            nutri_date.append(dateobj)
            nutri_calmorning.append(element['calmorning'])
            nutri_calafternoon.append(element['calafternoon'])
            nutri_calevening.append(element['calevening'])
            nutri_caltotal.append(element['caltotal'])
            # End of block

        fig, axs = plt.subplots(4)

        # Plot aerobic, anaerobic and steps
        axs[0].plot(nutri_date, nutri_calmorning)
        axs[1].plot(nutri_date, nutri_calafternoon, color='teal')
        axs[2].plot(nutri_date, nutri_calevening)
        axs[3].plot(nutri_date, nutri_caltotal)

        axs[0].set_ylabel('Morning', size=13)
        axs[1].set_ylabel('Afternoon', size=13)
        axs[2].set_ylabel('Evening', size=13)
        axs[3].set_ylabel('Total', size=13)

        fig.autofmt_xdate()
        fig.suptitle("Calories", size=20)
        holder = io.BytesIO()
        fig.savefig(holder, format="svg")
        decoded_holder = base64.b64encode(holder.getvalue()).decode()
        image = f"data:image/svg+xml;base64,{decoded_holder}"

        holder.close()
        return image

    def setNutrition(self, nutrition):
        self.current_nutrition = nutrition
        # Call the notifying signal
        self.nutritionChanged.emit()

    # SLEEP
    def read_sleep(self):
        # Retrieve the sleep history
        sleep = self.db.table('sleep')
        sleephist = sleep.all()
        return sleephist

    def getSleep(self):
        # Extracts the latest readings from Sleep table
        sleephist = self.read_sleep()
        # Init to empty string to avoid undefined val
        sleepobj = ['', '', '', '']
        if sleephist:
            sleep = sleephist[-1]  # Get the latest (newest) record

            dateobj = datefromisotz(sleep['timestamp'])
            date_repr = dateobj.strftime("%a, %b %d '%y - %H:%M")

            sleepobj = [str(date_repr), str(sleep['sleeptime']),
                        str(sleep['sleepquality'])]

        return sleepobj

    def sleepplot(self):
        # Retrieves all the history and packages into an array.
        sleephist = self.read_sleep()
        sleep_time = []
        sleep_date = []
        for element in sleephist:

            dateobj = datefromisotz(element['timestamp'])

            # Only print one value per day to avoid artifacts in plotting.
            sleep_date.append(dateobj)
            sleep_time.append(element['sleeptime'])
            # End of block

        fig = plt.figure()
        ax = fig.add_subplot(1, 1, 1)

        ax.plot(sleep_date, sleep_time, color="blue")

        ax.set_ylabel('Hours', size=13)
        fig.autofmt_xdate()
        fig.suptitle("Sleep hours", size=20)

        holder = io.BytesIO()
        fig.savefig(holder, format="svg")
        decoded_holder = base64.b64encode(holder.getvalue()).decode()
        image = f"data:image/svg+xml;base64,{decoded_holder}"

        holder.close()
        return image

    def setSleep(self, sleep):
        self.current_sleep = sleep
        # Call the notifying signal
        self.sleepChanged.emit()

    # PROPERTIES BLOCK
    # BP property to be accessed to and from QML and Python.
    # It is used in the context of showing the PA last results
    # in the main Lifestyle screen.
    pa = Property("QVariantList", getPA, setPA, notify=paChanged)

    # Property to retrieve the plot of the Physical Activity.
    paplot = Property(str, paplot, setPA, notify=paChanged)

    # Notifying signal - to be used in qml as "onPAChanged"
    nutritionChanged = Signal()

    # Nutrition property to be accessed to and from QML and Python.
    # It is used in the context of showing the Nutrition last results
    # in the main Lifestyle screen.
    nutrition = Property("QVariantList", getNutrition,
                         setNutrition, notify=nutritionChanged)

    # Property to retrieve the plot of the Physical Activity.
    nutritionplot = Property(str, nutriplot, setNutrition,
                             notify=nutritionChanged)

    # Notifying signal - to be used in qml as "onSleepChanged"
    sleepChanged = Signal()

    # Nutrition property to be accessed to and from QML and Python.
    # It is used in the context of showing the Nutrition last results
    # in the main Lifestyle screen.
    sleep = Property("QVariantList", getSleep,
                     setSleep, notify=sleepChanged)

    # Property to retrieve the plot of the Physical Activity.
    sleepplot = Property(str, sleepplot, setSleep,
                         notify=sleepChanged)



####################################################################
#   Copyright (C) 2020-2021 Luis Falcon <falcon@gnuhealth.org>
#   Copyright (C) 2020-2021 GNU Solidario <health@gnusolidario.org>
#   License: GPL v3+
#   Please read the COPYRIGHT and LICENSE files of the package
####################################################################

import datetime
from uuid import uuid4
from PySide2.QtCore import QObject, Signal, Slot
from tinydb import TinyDB
from mygnuhealth.myghconf import dbfile
from mygnuhealth.core import PageOfLife


class MoodEnergy(QObject):

    db = TinyDB(dbfile)

    def insert_values(self, mood, energy):
        mood_table = self.db.table('mood')
        current_date = datetime.datetime.now().isoformat()
        moodmon = False  # Init to false the mood monitoring process
        domain = 'medical'
        context = 'self_monitoring'

        if (energy > -1):  # Will evaluate to True since energy lower lim = 0
            moodmon = True
            mood_event_id = str(uuid4())
            synced = False
            mood_table.insert({'timestamp': current_date,
                               'event_id': mood_event_id,
                               'synced': synced,
                               'mood': mood,
                               'energy': energy})

            print("Saved Mood and Energy Levels", mood_event_id, synced, mood,
                  energy, current_date)

        if (moodmon):
            # This block is related to the Page of Life creation
            event_id = str(uuid4())
            monitor_readings = [
                {'mood_energy': {'mood': mood, 'energy': energy}},
                ]

            pol_vals = {
                'page': event_id,
                'page_date': current_date,
                'domain': domain,
                'context': context,
                'measurements': monitor_readings
                }

            # Create the Page of Life associated to this reading
            PageOfLife.create_pol(PageOfLife, pol_vals)

    @Slot(int, int)
    def getvals(self, *args):
        self.insert_values(*args)
        self.setOK.emit()

    # Signal to emit to QML if the mood and energy values were stored correctly
    setOK = Signal()



####################################################################
#   Copyright (C) 2020-2021 Luis Falcon <falcon@gnuhealth.org>
#   Copyright (C) 2020-2021 GNU Solidario <health@gnusolidario.org>
#   License: GPL v3+
#   Please read the COPYRIGHT and LICENSE files of the package
####################################################################

import requests


def test_federation_connection(protocol, host, port, acct, passwd):
    """
    Connection test to Thalamus Server
    """
    conn = ''

    url = f'{protocol}://{host}:{port}/people/{acct}'

    try:
        conn = requests.get(url, auth=(acct, passwd), verify=False)
        # If there is an error, raise the exception
        conn.raise_for_status()

    # Connection / Network errors
    except requests.exceptions.ConnectionError as e:
        print(f"Connection error: {e}")
        login_status = -2

    # Catch HTTP exceptions, including unauthorized (401) and (404)
    except requests.exceptions.HTTPError as e:
        print(f"HTPP error: {e}")
        login_status = -1

    if conn:
        print("***** Connection to Thalamus Server OK !******")
        login_status = 0

    return login_status




####################################################################
#   Copyright (C) 2020-2021 Luis Falcon <falcon@gnuhealth.org>
#   Copyright (C) 2020-2021 GNU Solidario <health@gnusolidario.org>
#   License: GPL v3+
#   Please read the COPYRIGHT and LICENSE files of the package
####################################################################

import datetime
from uuid import uuid4
from PySide2.QtCore import QObject, Signal, Slot
from tinydb import TinyDB
from mygnuhealth.myghconf import dbfile
from mygnuhealth.core import PageOfLife


class GHSleep(QObject):

    db = TinyDB(dbfile)

    def insert_values(self, sleeptime, sleepquality, information):
        sleep = self.db.table('sleep')
        current_date = datetime.datetime.now().isoformat()
        domain = 'lifestyle'
        context = 'sleep'

        sleep_event_id = str(uuid4())
        synced = False
        sleep.insert({'timestamp': current_date,
                      'event_id': sleep_event_id,
                      'synced': synced,
                      'sleeptime': sleeptime,
                      'sleepquality': sleepquality})

        print("Saved Sleep information", sleep_event_id,
              sleeptime, sleepquality, current_date)

        # Page of Life block related to Nutrition
        event_id = str(uuid4())
        monitor_readings = [
            {'sleep': {'sleeptime': sleeptime,
                       'sleepquality': sleepquality}},
            ]

        pol_vals = {
            'page': event_id,
            'page_date': current_date,
            'domain': domain,
            'context': context,
            'measurements': monitor_readings,
            'info': information
            }

        # Create the Page of Life associated to this reading
        PageOfLife.create_pol(PageOfLife, pol_vals)

    @Slot(int, str, str)
    def getvals(self, sleeptime, sleepquality, information):
        self.insert_values(sleeptime, sleepquality, information)
        self.setOK.emit()

    # Signal to emit to QML if the values were stored correctly
    setOK = Signal()



####################################################################
#   Copyright (C) 2020-2021 Luis Falcon <falcon@gnuhealth.org>
#   Copyright (C) 2020-2021 GNU Solidario <health@gnusolidario.org>
#   License: GPL v3+
#   Please read the COPYRIGHT and LICENSE files of the package
####################################################################

import json
import requests
import threading

from PySide2.QtCore import QObject, Signal, Slot, Property
from tinydb import TinyDB, Query
from mygnuhealth.myghconf import bolfile, dbfile
from mygnuhealth.core import datefromisotz


class GHBol(QObject):
    """Class that manages the person Book of Life

        Attributes:
        -----------
            boldb: TinyDB instance.
                Holds the book of life with all the events (pages of life)
        Methods:
        --------
            read_book: retrieves all pages
            format_bol: compacts and shows the relevant fields in a
            human readable format
    """

    boldb = TinyDB(bolfile)
    db = TinyDB(dbfile)

    def format_bol(self, bookoflife):
        """Takes the necessary fields and formats the book in a way that can
        be shown in the device, mixing fields and compacting entries in a more
        human readable format"""
        book = []
        for pageoflife in bookoflife:
            pol = {}
            dateobj = datefromisotz(pageoflife['page_date'])
            # Use a localized and easy to read date format
            date_repr = dateobj.strftime("%a, %b %d '%y\n%H:%M")

            pol['date'] = date_repr
            pol['domain'] = (f"{pageoflife['domain']} "
                             f"({pageoflife['context']})")

            summ = ''
            msr = ''

            title = pageoflife['summary']
            details = pageoflife['info']

            mvals = pageoflife['measurements']

            if title:
                summ = f'{title}\n'

            if ('measurements' in pageoflife.keys() and
                    mvals):
                measure_d = {
                    "bg": ("Blood glucose", "mg/dl"),
                    "hr": ("Heart rate", "bpm"),
                    "wt": ("Weight", "kg"),
                    "bmi": ("BMI", "kg/m2"),
                    "osat": ("osat", "%"),
                }

                for measure in mvals:
                    measure_keys = measure.keys()

                    for key, value in measure_d.items():
                        if key in measure_keys:
                            msr = (f"{msr}{value[0]}: {measure[key]} "
                                   f"{value[1]}<br/>")

                if 'bp' in mvals[0].keys():

                    msr = (f"{msr}"
                           f"BP: {mvals[0]['bp']['systolic']} / "
                           f"{mvals[0]['bp']['diastolic']} mmHg<br/>")

                if 'mood_energy' in mvals[0].keys():
                    msr = (f"{msr}"
                           f"mood: {mvals[0]['mood_energy']['mood']}<br/>"
                           f"energy: {mvals[0]['mood_energy']['energy']}<br/>")

                summ = summ + msr

                # Include the Lifestyle measures
                if (pageoflife['domain'] == 'lifestyle'):
                    measurements = mvals[0]
                    measurements_keys = measurements.keys()

                    # Show / format the Physical activity values ("pa" key)
                    if 'pa' in measurements_keys:
                        for key, value in measurements['pa'].items():
                            summ = f"{summ}{key}: {value}<br/>"

                    # Show / format the nutrition values ("nutrition" key)
                    if 'nutrition' in measurements_keys:
                        for key, value in measurements['nutrition'].items():
                            summ = f"{summ}{key}: {value}<br/>"

                    # Show / format the sleep values ("sleep" key)
                    if 'sleep' in measurements_keys:
                        for key, value in measurements['sleep'].items():
                            summ = f"{summ}{key}: {value}<br/>"

            if ('genetic_info' in pageoflife.keys() and
                    pageoflife['genetic_info']):
                genetics = pageoflife['genetic_info']
                summ = f'{summ}{genetics}<br/>'

            if details:
                summ = f'{summ}{details}<br/>'

            pol['summary'] = summ
            book.append(pol)
        return book

    def read_book(self):
        """retrieves all pages of the individual Book of Life
        """
        booktable = self.boldb.table('pol')
        book = booktable.all()
        formatted_bol = self.format_bol(book)
        return formatted_bol

    def check_sync_status(self):
        fedinfo = self.db.table('federation')
        if len(fedinfo):
            sync = fedinfo.all()[0]['enable_sync']
            return sync

    @Slot(str)
    def sync_book(self, fedkey):
        # Emit the signal to display a busy indicator while
        # pushing the pages of life to the GH federation
        print("***** Initiating the synchronization \
                with the GH federation server")
        self.pushingPols.emit()
        # Spawn a new thread so synchronization / pushing is done asynchronously
        # in a non-blocking fashion
        thread = threading.Thread(name="pushpols_thread",
                                  target=self.push_pols, args=(fedkey,))
        thread.start()

    def push_pols(self, fedkey):
        """This method will go through each page in the book of life
        that has not been sent to the GNU Health Federation server yet
        (fsynced = False).
        It also checks for records that have a book associated to it
        and that the specific page is has not the "private" flag set.

        Parameters
        ----------
        """
        fedinfo = self.db.table('federation')
        if len(fedinfo):
            res = fedinfo.all()[0]

        # Refresh all pages of life
        booktable = self.boldb.table('pol')
        book = booktable.all()
        user = res['federation_account']
        protocol = res['protocol']
        server = res['federation_server']
        port = res['federation_port']

        for pol in book:
            timestamp = pol['page_date']
            node = pol['node']
            page_id = pol['page']
            synced = pol['fsynced']

            # Only sync those pages that are not private
            if 'privacy' in pol.keys():
                privacy = pol['privacy']
                if not privacy and not synced:
                    creation_info = {'user': user, 'timestamp': timestamp,
                                     'node': node}

                    pol['creation_info'] = creation_info
                    pol['id'] = page_id

                    url = f"{protocol}://{server}:{port}/pols/{user}/{page_id}"

                    pol['fsynced'] = True
                    send_data = requests.request('POST', url,
                                                 data=json.dumps(pol),
                                                 auth=(user, fedkey),
                                                 verify=False)
                    if send_data:
                        print("Page successfully sent to the GH Federation",
                              send_data.status_code)
                        # Update page of life sync status locally to true
                        print("Setting fsynced to True on page... ", page_id)
                        Page = Query()
                        booktable.update({'fsynced': True},
                                         Page.page == page_id)

                    else:
                        print("Error sending the page.",
                              send_data.status_code)

                else:
                    if privacy:
                        print("This page is private, not syncing", pol)
                    if synced:
                        print("Page already synced", pol)

        # Emit the signal to remove the busy indicator
        print("***** Done pushing PoLs... disabling busy indicator")
        self.finishSyncPols.emit()

    # Signal to emit when the pages of life are being sent to the Federation
    pushingPols = Signal()

    # Signal to emit when the pages of life have been sent
    finishSyncPols = Signal()

    # Property block
    book = Property("QVariantList", read_book, constant=True)

    # Expose to QML the value of sync status
    # It will disable the password field if sync is not enabled.
    sync_status = Property(bool, check_sync_status, constant=True)



####################################################################
#   Copyright (C) 2020-2021 Luis Falcon <falcon@gnuhealth.org>
#   Copyright (C) 2020-2021 GNU Solidario <health@gnusolidario.org>
#   License: GPL v3+
#   Please read the COPYRIGHT and LICENSE files of the package
####################################################################

import io
from PySide2.QtCore import QObject, Signal, Property
from tinydb import TinyDB
import matplotlib.pyplot as plt
import base64
from mygnuhealth.core import datefromisotz
from mygnuhealth.myghconf import dbfile


class GHPsycho(QObject):

    def __init__(self):
        QObject.__init__(self)

        self.current_mood = ""
        self.current_energy = ""

    db = TinyDB(dbfile)

    def read_mood(self):
        # Retrieve the  history
        mood_table = self.db.table('mood')
        moodhist = mood_table.all()
        return moodhist

    def getMood(self):
        # Extracts the latest readings from mood table (both mood and energy)
        moodhist = self.read_mood()
        moodobj = ['', '', '']  # Init to empty string to avoid undefined val
        if moodhist:
            mood = moodhist[-1]  # Get the latest (newest) record

            dateobj = datefromisotz(mood['timestamp'])
            date_repr = dateobj.strftime("%a, %b %d '%y - %H:%M")

            moodobj = [str(date_repr), str(mood['mood']), str(mood['energy'])]

        return moodobj

    def moodplot(self):
        # Retrieves all the history and packages into an array.
        moodhist = self.read_mood()
        mood = []
        energy = []
        mood_date = []
        for element in moodhist:

            dateobj = datefromisotz(element['timestamp'])

            # Only print one value per day to avoid artifacts in plotting.
            mood_date.append(dateobj)
            mood.append(element['mood'])
            energy.append(element['energy'])
            # end block

        print(f"Info to plot...{mood} {energy} {moodhist}")

        fig, axs = plt.subplots(2)

        # Plot both mood and energy history
        axs[0].plot(mood_date, mood)
        axs[1].plot(mood_date, energy, color='teal')

        axs[0].set_ylabel('Mood', size=13)
        axs[1].set_ylabel('Energy', size=13)

        axs[0].set_ylim([-3, 3])
        axs[1].set_ylim([0, 3])

        fig.autofmt_xdate()
        fig.suptitle("Mood and Energy", size=20)
        holder = io.BytesIO()
        fig.savefig(holder, format="svg")
        decoded_holder = base64.b64encode(holder.getvalue()).decode()
        image = f"data:image/svg+xml;base64,{decoded_holder}"

        holder.close()
        return image

    def setMood(self, mood):
        self.current_mood = mood
        # Call the notifying signal
        self.moodChanged.emit()

    # PROPERTIES BLOCK
    # Notifying signal - to be used in qml as "onMoodChanged"
    moodChanged = Signal()

    # mood property to be accessed to and from QML and Python.
    # It is used in the context of showing the Mood and Energy last results
    # in the main bio screen.
    mood = Property("QVariantList", getMood, setMood, notify=moodChanged)

    # Property to retrieve the plot of the Mood and Energy
    moodplot = Property(str, moodplot, setMood, notify=moodChanged)



####################################################################
#   Copyright (C) 2020-2021 Luis Falcon <falcon@gnuhealth.org>
#   Copyright (C) 2020-2021 GNU Solidario <health@gnusolidario.org>
#   License: GPL v3+
#   Please read the COPYRIGHT and LICENSE files of the package
####################################################################

import datetime
from uuid import uuid4
from PySide2.QtCore import QObject, Signal, Slot
from tinydb import TinyDB
from mygnuhealth.myghconf import dbfile
from mygnuhealth.core import PageOfLife


class GHNutrition(QObject):

    db = TinyDB(dbfile)

    def insert_values(self, calmorning, calafternoon, calevening, caltotal,
                      information):
        nutrition = self.db.table('nutrition')
        current_date = datetime.datetime.now().isoformat()
        domain = 'lifestyle'
        context = 'nutrition'

        nutrition_event_id = str(uuid4())
        synced = False
        nutrition.insert({'timestamp': current_date,
                          'event_id': nutrition_event_id,
                          'synced': synced,
                          'calmorning': calmorning,
                          'calafternoon': calafternoon,
                          'calevening': calevening,
                          'caltotal': caltotal})

        print("Saved Nutrition information", nutrition_event_id,
              calmorning, calafternoon, calevening, caltotal, current_date)

        # Page of Life block related to Nutrition
        event_id = str(uuid4())
        monitor_readings = [
            {'nutrition': {'calmorning': calmorning,
                           'calafternoon': calafternoon,
                           'calevening': calevening, 'caltotal': caltotal}},
            ]

        pol_vals = {
            'page': event_id,
            'page_date': current_date,
            'domain': domain,
            'context': context,
            'measurements': monitor_readings,
            'info': information
            }

        # Create the Page of Life associated to this reading
        PageOfLife.create_pol(PageOfLife, pol_vals)

    @Slot(int, int, int, int, str)
    def getvals(self, calmorning, calafternoon, calevening, caltotal,
                information):
        self.insert_values(calmorning, calafternoon, calevening, caltotal,
                           information)
        self.setOK.emit()

    # Signal to emit to QML if the blood pressure values were stored correctly
    setOK = Signal()



####################################################################
#   Copyright (C) 2020-2021 Luis Falcon <falcon@gnuhealth.org>
#   Copyright (C) 2020-2021 GNU Solidario <health@gnusolidario.org>
#   Copyright (C) 2020-2021 Carl Schwan <carlschwan@kde.org>
#   License: GPL v3+
#   Please read the COPYRIGHT and LICENSE files of the package
####################################################################

from PySide2.QtCore import QObject, Signal, Slot, Property
import bcrypt
from tinydb import TinyDB
from mygnuhealth.myghconf import dbfile
from mygnuhealth.core import get_personal_key, check_date, get_user_profile

import logging
import datetime


class LocalAccountManager(QObject):
    """This class manages the login and initialization of the personal key

        Attributes:
        -----------
            loginSuccess: Signal to emit when entering the correct personal key
            invalidCredentials: Signal to emit when entering the wrong key
            wrongDate: Signal to emit when an invalid date is found

            accountExist: Property that checks if the user account
                has been created
            todayDate: Property with current date

        Methods:
        --------
            init_personal_key: Sets the personal key at the initial run.
            login: Slot that receives the personal key to login, and checks
                    if it is the correct one to log in.
            createAccount: Slot that receives the initial personal key to
                create an account.

    """

    def __init__(self):
        QObject.__init__(self)
        self.db = TinyDB(dbfile)

    def get_date(self):
        """
        Returns the date packed into an array (day,month,year)
        """
        rightnow = datetime.datetime.now()
        dateobj = []
        dateobj.append(rightnow.day)
        dateobj.append(rightnow.month)
        dateobj.append(rightnow.year)
        return dateobj

    def account_exist(self):
        """
        Check if an account exist in the database.
        """
        if (self.db.table('credentials')):
            print("DB is initialized")
            rc = True

        else:
            print("We need to init the personal Key")
            rc = False

        return rc

    def update_profile(self, profile):
        # Include height and other user settings at initialization
        profiletable = self.db.table('profile')
        if (len(profiletable) > 0):
            print(f"Updating profile ... {profile}")
            profiletable.update(profile)

        else:
            print(f"Initializing profile. Setting profile {profile}")
            profiletable.insert(profile)

    def init_personal_key(self, key):
        encrypted_key = bcrypt.hashpw(key.encode('utf-8'),
                                      bcrypt.gensalt()).decode('utf-8')

        credentialstable = self.db.table('credentials')
        if (len(credentialstable) > 0):
            credentialstable.update({'personal_key': encrypted_key})
        else:
            logging.info("Initializing credentials table")
            credentialstable.insert({'personal_key': encrypted_key})

        logging.info(f"Initialized personal key: {encrypted_key}")
        return encrypted_key

    def get_username(self):
        if get_user_profile(self.db):
            return get_user_profile(self.db)['username']

    @Slot(str)
    def login(self, key):
        key = key.encode()

        personal_key = get_personal_key(self.db)

        if bcrypt.checkpw(key, personal_key):
            logging.info("Login correct - Move to main PHR page")
            self.loginSuccess.emit()
        else:
            self.invalidCredentials.emit()

    @Slot(str, str, str, list)
    def createAccount(self, key, height, personname, birthdate):
        # Retrieves the information from the initialization form
        # Initializes user profile and sets the initial password
        validation_process = True
        if (height):
            # Sets the user height
            height = int(height)
            profile = {'height': height}
            self.update_profile(profile)

        if (personname):
            # Sets the user name
            profile = {'username': personname}
            self.update_profile(profile)

        if (birthdate):
            if (check_date(birthdate)):
                # Sets the birthdate
                year, month, day = birthdate
                daterp = str(datetime.date(year, month, day))
                profile = {'birthdate': daterp}
                self.update_profile(profile)
            else:
                print("Wrong Date!")
                validation_process = False
                self.wrongDate.emit()

        key = key.encode()
        if (validation_process and
                self.init_personal_key(key.decode('utf-8'))):
            self.loginSuccess.emit()

    # Properties
    accountExist = Property(bool, account_exist, constant=True)

    todayDate = Property("QVariantList", get_date, constant=True)

    person = Property(str, get_username, constant=True)

    # Signals
    loginSuccess = Signal()
    invalidCredentials = Signal()
    wrongDate = Signal()



#!/usr/bin/env python3
# This file if part of the MyGNUHealth, the GNU Health PHR
# www.gnuhealth.org

from setuptools import setup, find_packages

long_desc = open('README.rst').read()

# Initialize about
about = {}
with open("mygnuhealth/about.py") as fp:
    exec(fp.read(), about)

setup(
    name=about['__appname__'],
    version=about['__version__'],
    description=about['__description__'],
    url=about['__homepage__'],
    license=about['__license__'],
    author=about['__author__'],
    author_email=about['__email__'],
    download_url=about['__download_url__'],
    long_description=long_desc,
    keywords='eHealth PHIMS PHR mHealth GNUHealth',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: X11 Applications :: KDE',
        'Environment :: X11 Applications :: Qt',
        'Topic :: Desktop Environment :: K Desktop Environment (KDE)',
        'Intended Audience :: Developers',
        'Intended Audience :: End Users/Desktop',
        'Intended Audience :: Healthcare Industry',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: GNU General Public License (GPL)',
        'Operating System :: POSIX',
        'Programming Language :: Python :: 3',
        'Topic :: Scientific/Engineering :: Bio-Informatics',
        'Topic :: Scientific/Engineering :: Medical Science Apps.',
        'Natural Language :: English'
        ],
    platforms='any',
    scripts=['bin/mygnuhealth'],
    python_requires='>=3.6,<4',
    install_requires=[
        'matplotlib',
        'requests',
        'tinydb',
        'bcrypt'
        ],
    packages=find_packages(),
    include_package_data=True,
    zip_safe=False,
    data_files=[
        ('share/applications', ['org.kde.mygnuhealth.desktop']),
        ('share/metainfo', ['org.kde.mygnuhealth.metainfo.xml']),
        ('share/icons/hicolor/scalable/apps/',
         ['mygnuhealth/images/mygnuhealth.svg']),
        ]
      )
